from opentsdb import TSDBClient, Counter, Gauge


def test_count(http_client2: TSDBClient):
    http_client2.is_connected()
    http_client2.TEST_METRIC = Counter('test.metric.counter')
    metric = http_client2.TEST_METRIC.inc()
    assert metric['value'] == 1
    metric = http_client2.TEST_METRIC.inc()
    assert metric['value'] == 2


def test_count_with_tag(http_client2: TSDBClient):
    http_client2.is_connected()
    http_client2.TEST_METRIC2 = Counter('test.metric2.counter', ['tag1'])
    metric = http_client2.TEST_METRIC2.tags('tag1').inc()
    assert 'tag1' in metric['tags']


def test_gauge(http_client2: TSDBClient):
    http_client2.is_connected()
    http_client2.TEST_METRIC = Gauge('test.metric.gauge')
    metric = http_client2.TEST_METRIC.inc()
    assert metric['value'] == 1
    metric = http_client2.TEST_METRIC.set(10)
    assert metric['value'] == 10
    metric = http_client2.TEST_METRIC.dec()
    assert metric['value'] == 9



import threading
import logging

logger = logging.getLogger('opentsdb-py')


class TSDBConnect:

    def __init__(self, host: str, port: int, check_tsdb_alive: bool=False, *args, **kwargs):
        self.tsdb_host = host
        self.tsdb_port = int(port)
        self._connect = None

        if check_tsdb_alive:
            self.is_alive(raise_error=True)

        self.stopped = threading.Event()

    def is_alive(self, timeout=3, raise_error=False) -> bool:
        raise NotImplementedError()

    @property
    def connect(self):
        raise NotImplementedError()

    def disconnect(self):
        logger.debug("Disconnecting...")
        self.stopped.set()
        if self._connect:
            self._connect.close()
        self._connect = None

    def sendall(self, metric: dict):
        raise NotImplementedError()



from os import environ

import pytest

from opentsdb import TSDBClient, TSDBConnectProtocols, Counter


@pytest.fixture
def tsdb_host():
    return environ.get('OPEN_TSDB_HOST', '127.0.0.1')


@pytest.fixture
def tsdb_port():
    return int(environ.get('OPEN_TSDB_PORT', '4242'))


@pytest.fixture
def http_client(tsdb_host, tsdb_port):
    return TSDBClient(tsdb_host, tsdb_port, host_tag=False)


@pytest.fixture
def http_client2(tsdb_host, tsdb_port):
    return TSDBClient(tsdb_host, tsdb_port, host_tag=True)


class Metrics(TSDBClient):
    PREDEFINED_METRIC = Counter('test.predefined.metric')


@pytest.fixture
def http_client3(tsdb_host, tsdb_port):
    return Metrics(tsdb_host, tsdb_port, host_tag=True)


@pytest.fixture
def telnet_client(tsdb_host, tsdb_port):
    return TSDBClient(tsdb_host, tsdb_port, protocol=TSDBConnectProtocols.TELNET, host_tag=True)



import logging
import socket
import time

from opentsdb.protocols.tsdb_connect import TSDBConnect
from opentsdb.exceptions import TSDBNotAlive

logger = logging.getLogger('opentsdb-py')


class TelnetTSDBConnect(TSDBConnect):

    def is_alive(self, timeout=3, raise_error=False) -> bool:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(timeout)
            sock.connect((self.tsdb_host, self.tsdb_port))
            sock.close()
        except (ConnectionRefusedError, socket.timeout) as error:
            if raise_error:
                raise TSDBNotAlive(str(error))
            return False
        else:
            return True

    @property
    def connect(self) -> socket.socket:
        if not self._connect or getattr(self._connect, '_closed', False) is True:
            logger.debug("Connect to OpenTSDB: %s:%s", self.tsdb_host, self.tsdb_port)
            self.stopped.clear()
            self._make_connection()

        return self._connect

    def _make_connection(self, timeout=2):
        attempt = 0
        while not self.stopped.is_set():
            try:
                self._connect = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self._connect.settimeout(timeout)
                self._connect.connect((self.tsdb_host, self.tsdb_port))
                return
            except (ConnectionRefusedError, socket.timeout):
                time.sleep(min(15, 2 ** attempt))
                attempt += 1

    def sendall(self, metric: dict):
        try:
            tags_string = ' '.join(['%s=%s' % (key, value) for key, value in metric['tags'].items()])
            metric_str = "put %s %d %s %s\n" % (metric['metric'], metric['timestamp'], metric['value'], tags_string)
            logger.debug("Send metric: %s", metric_str)
            self.connect.sendall(metric_str.encode('utf-8'))
        except Exception:
            self._connect.close()
            raise



import time

import pytest

from opentsdb.tsdb_client import TSDBClient, TSDBConnectProtocols
from opentsdb.exceptions import ValidationError, TSDBNotAlive

MAX_STOP_DURATION = 1  # seconds


def test_no_tag_send(http_client: TSDBClient):
    pytest.raises(ValidationError, http_client.send, 'test', 1)


def test_bad_value_send(http_client: TSDBClient):
    pytest.raises(ValidationError, http_client.send, 'test', b'')
    pytest.raises(ValidationError, http_client.send, 'test', [])
    pytest.raises(ValidationError, http_client.send, 'test', {})


def test_bad_chars_send(http_client: TSDBClient):
    pytest.raises(ValidationError, http_client.send, '%test%', 1)


def test_host_tag(telnet_client: TSDBClient):
    metric = telnet_client.send('test', 1)
    assert 'host' in metric['tags']


def test_check_tsdb_alive():
    pytest.raises(
        TSDBNotAlive,
        TSDBClient, '127.0.0.2', 42424, protocol=TSDBConnectProtocols.TELNET, check_tsdb_alive=True)


def test_http_client_close(http_client: TSDBClient):
    start = time.time()
    assert http_client.is_connected()
    http_client.close()
    http_client.wait()
    assert http_client.is_connected() is False
    assert time.time() - start < MAX_STOP_DURATION


def test_telnet_client_close(telnet_client: TSDBClient):
    start = time.time()
    assert telnet_client.is_connected()
    telnet_client.close()
    telnet_client.wait()
    assert telnet_client.is_connected() is False
    assert time.time() - start < MAX_STOP_DURATION


def test_send_metric_through_http(http_client: TSDBClient):
    _test_sending_metric(http_client)


def test_send_metric_through_telnet(telnet_client: TSDBClient):
    _test_sending_metric(telnet_client)


def _test_sending_metric(client: TSDBClient):
    client.send('test', 1, tag1=1)
    assert client.queue_size() == 1
    client.close()
    client.wait()
    assert client.queue_size() == 0
    assert client.statuses['success'] == 1


def test_predefined_metrics(http_client3: TSDBClient):
    assert http_client3.PREDEFINED_METRIC.client is not None



import logging
import queue
import socket
import string
import threading
import time
from os import environ
from typing import Optional

from opentsdb.exceptions import ValidationError
from opentsdb.metrics import Metric
from opentsdb.protocols import TSDBConnectProtocols

logger = logging.getLogger('opentsdb-py')


class TSDBClient:
    TSDB_HOST = environ.get('OPEN_TSDB_HOST', '127.0.0.1')
    TSDB_PORT = int(environ.get('OPEN_TSDB_PORT', 4242))
    TSDB_URI = environ.get('OPEN_TSDB_URI')
    TSDB_PUT_ENDPOINT = environ.get('OPEN_TSDB_PUT_ENDPOINT', '/api/put?details=true')
    TSDB_VERSION_ENDPOINT = environ.get('OPEN_TSDB_VERSION_ENDPOINT', '/api/version')

    TSDB_MAX_METRICS_QUEUE_SIZE = int(environ.get('TSDB_MAX_METRICS_QUEUE_SIZE', 10000))
    TSDB_SEND_METRICS_PER_SECOND_LIMIT = int(environ.get('TSDB_SEND_METRICS_PER_SECOND_LIMIT', 1000))
    TSDB_SEND_METRICS_BATCH_LIMIT = int(environ.get('TSDB_SEND_METRICS_BATCH_LIMIT', 50))
    TSDB_DEFAULT_HTTP_COMPRESSION = environ.get('TSDB_DEFAULT_HTTP_COMPRESSION', 'gzip')
    VALID_METRICS_CHARS = set(string.ascii_letters + string.digits + '-_./')

    def __init__(
            self,
            host: str = TSDB_HOST,
            port: int = TSDB_PORT,
            check_tsdb_alive: bool = False,
            protocol: str = TSDBConnectProtocols.HTTP,
            run_at_once: bool = True,
            static_tags: dict = None,
            host_tag: bool = True,
            max_queue_size: int = TSDB_MAX_METRICS_QUEUE_SIZE,
            http_compression: str = TSDB_DEFAULT_HTTP_COMPRESSION,
            uri: str = TSDB_URI,
            send_metrics_limit: int = TSDB_SEND_METRICS_PER_SECOND_LIMIT,
            send_metrics_batch_limit: int = TSDB_SEND_METRICS_BATCH_LIMIT,
            put_endpoint: str = TSDB_PUT_ENDPOINT,
            version_endpoint: str = TSDB_VERSION_ENDPOINT,
    ):

        self.host_tag = host_tag
        self.protocol = protocol
        self.check_tsdb_alive = check_tsdb_alive
        self.static_tags = static_tags or {}
        self.send_metrics_limit = send_metrics_limit if protocol == TSDBConnectProtocols.TELNET else 0
        self.send_metrics_batch_limit = send_metrics_batch_limit if protocol == TSDBConnectProtocols.HTTP else 0
        self.http_compression = http_compression

        self._tsdb_connect = None
        self._close_client = threading.Event()
        self._metrics_queue = queue.Queue(maxsize=max_queue_size)
        self.statuses = {'success': 0, 'failed': 0, 'queued': 0}

        self._metric_send_thread = None

        if run_at_once is True:
            self.init_client(host, port, uri, put_endpoint, version_endpoint)

    def init_client(
            self, host, port: int = TSDB_PORT, uri: Optional[str] = None,
            put_endpoint: str = TSDB_PUT_ENDPOINT,
            version_endpoint: str = TSDB_VERSION_ENDPOINT
    ):
        self._tsdb_connect = TSDBConnectProtocols.get_connect(
            self.protocol, host, port, self.check_tsdb_alive,
            compression=self.http_compression, uri=uri,
            put_endpoint=put_endpoint, version_endpoint=version_endpoint
        )

        self._metric_send_thread = TSDBConnectProtocols.get_push_thread(
            self.protocol, self._tsdb_connect, self._metrics_queue, self._close_client,
            self.send_metrics_limit, self.send_metrics_batch_limit, self.statuses)
        self._metric_send_thread.daemon = True
        self._metric_send_thread.start()

        self._load_predefined_metrics()

    def _load_predefined_metrics(self):
        for key, value in self.__class__.__dict__.items():
            if isinstance(value, Metric):
                self.__setattr__(key, value)

    def __setattr__(self, key, value):
        if isinstance(value, Metric):
            if value.client is None:
                value.client = self

        super(TSDBClient, self).__setattr__(key, value)

    def is_connected(self) -> bool:
        return self._metric_send_thread.is_alive()

    def is_alive(self) -> bool:
        return self._tsdb_connect.is_alive()

    def close(self, force=False):
        self._close_client.set()
        self._metrics_queue.put(StopIteration)
        if force and self._tsdb_connect:
            self._tsdb_connect.stopped.set()

    def wait(self):
        while self.is_connected():
            time.sleep(0.05)

    def queue_size(self) -> int:
        return self._metrics_queue.qsize()

    def send(self, name: str, value, **tags) -> dict:
        tags.update(self.static_tags)
        if self.host_tag is True and 'host' not in tags:
            tags['host'] = socket.gethostname()

        self._validate_metric(name, value, tags)
        metric = dict(metric=name, timestamp=int(tags.pop('timestamp', time.time())), value=value, tags=tags)

        if not self._close_client.is_set():
            self._push_metric_to_queue(metric)

        return metric

    def _validate_metric(self, name, value, tags):
        try:
            assert all(char in self.VALID_METRICS_CHARS for char in name), \
                "Metric name contain incorrect chars '%s'" % name
            assert isinstance(value, (str, int, float)), "Incorrect metric value type '%s'" % type(value)
            assert tags != {}, "Need at least one tag"
        except AssertionError as error:
            raise ValidationError("Metric not valid: %s" % str(error))

    def _push_metric_to_queue(self, metric):
        try:
            self._metrics_queue.put(metric, False)
        except queue.Full:
            logger.warning("Drop oldest metric because Queue is full.")
            self._metrics_queue.get()
            self._metrics_queue.put(metric, False)

        self.statuses['queued'] += 1



from logging import getLogger
from queue import Empty
import threading
import random
import time

logger = getLogger('opentsdb-py')


class PushThread(threading.Thread):

    WAIT_NEXT_METRIC_TIMEOUT = 3

    def __init__(self, tsdb_connect, metrics_queue, close_client,
                 send_metrics_limit, send_metrics_batch_limit, statuses):
        super().__init__()
        self.tsdb_connect = tsdb_connect
        self.metrics_queue = metrics_queue
        self.close_client_flag = close_client
        self.send_metrics_limit = send_metrics_limit
        self.send_metrics_batch_limit = send_metrics_batch_limit
        self.statuses = statuses

        self._retry_send_metrics = None

    def run(self):
        while not self._is_done():
            start_time = time.time()

            try:
                if self._retry_send_metrics:
                    data = self._retry_send_metrics
                    self._retry_send_metrics = None
                else:
                    data = self._next(self.WAIT_NEXT_METRIC_TIMEOUT)

                self.send(data)
            except StopIteration:
                break
            except Empty:
                continue
            except Exception as error:
                logger.exception(error)

            if self.send_metrics_limit > 0:
                self.__metrics_limit_timeout(start_time)

        self.tsdb_connect.disconnect()

    def _is_done(self):
        return self.tsdb_connect.stopped.is_set() or (self.close_client_flag.is_set() and self.metrics_queue.empty())

    def _next(self, wait_timeout):
        raise NotImplementedError()

    def send(self, data):
        raise NotImplementedError()

    def __metrics_limit_timeout(self, start_time):
        pass

    def _update_statuses(self, success, failed):
        self.statuses['success'] += success
        self.statuses['failed'] += failed


class HTTPPushThread(PushThread):

    def _next(self, wait_timeout):
        total_metrics = self.metrics_queue.qsize()
        iter_count = total_metrics if total_metrics <= self.send_metrics_batch_limit else self.send_metrics_batch_limit
        metrics = []
        if total_metrics:
            for _ in range(iter_count):
                metrics.append(self.metrics_queue.get_nowait())
        else:
            metrics.append(self.metrics_queue.get(block=True, timeout=wait_timeout))

        if StopIteration in metrics and len(metrics) == 1:
            raise StopIteration
        elif StopIteration in metrics:
            metrics.remove(StopIteration)
            self.metrics_queue.put(StopIteration)
        return metrics

    def send(self, data):
        try:
            result = self.tsdb_connect.sendall(*data)
        except Exception as error:
            logger.exception("Push metric failed: %s", error)
            self._retry_send_metrics = data
            time.sleep(1)
        else:
            failed = result.get('failed', 0)
            self._update_statuses(result.get('success', 0), failed)
            if failed:
                logger.warning("Push metrics are failed %d/%d" % (failed, len(data)),
                               extra={'errors': result.get('errors')})


class TelnetPushThread(PushThread):

    def _next(self, wait_timeout):
        metric = self.metrics_queue.get(block=True, timeout=wait_timeout)

        if metric is StopIteration:
            raise metric
        return metric

    def __metrics_limit_timeout(self, start_time):
        duration = time.time() - start_time
        wait_time = (2.0 * random.random()) / self.send_metrics_limit
        if wait_time > duration:
            logger.debug("Wait for %s", wait_time - duration)
            time.sleep(wait_time - duration)

    def send(self, data):
        try:
            self.tsdb_connect.sendall(data)
        except Exception as error:
            logger.exception("Push metric failed: %s", error)
            self._retry_send_metrics = data
            time.sleep(1)
        else:
            self._update_statuses(1, 0)



import logging
import json
import gzip
import os.path
from typing import Optional

from requests import Session

from opentsdb.protocols.tsdb_connect import TSDBConnect
from opentsdb.exceptions import TSDBNotAlive

logger = logging.getLogger('opentsdb-py')


class TSDBUrls:

    def __init__(self, base_url, version_endpoint, put_endpoint):
        self.version = base_url + version_endpoint
        self.put = base_url + put_endpoint

    @classmethod
    def from_host_and_port(cls, host, port, version_endpoint, put_endpoint):
        base_url = 'http://{}:{}'.format(host, port)
        return cls(base_url, version_endpoint, put_endpoint)

    @classmethod
    def from_uri(cls, uri, version_endpoint, put_endpoint):
        return cls(uri, version_endpoint, put_endpoint)


class HttpTSDBConnect(TSDBConnect):
    SEND_TIMEOUT = 2

    def __init__(self, host: str, port: int, check_tsdb_alive: bool,
                 compression: str, version_endpoint: str, put_endpoint: str, uri: Optional[str]):
        if uri is None:
            self.tsdb_urls = TSDBUrls.from_host_and_port(host, int(port), version_endpoint, put_endpoint)
        else:
            self.tsdb_urls = TSDBUrls.from_uri(uri, version_endpoint, put_endpoint)
        super().__init__(host, port, check_tsdb_alive)
        self.compression = compression
        assert self.compression in ['gzip', None], 'Unsupported HTTP compression type: %s' % self.compression
        if self.compression:
            logger.info("Compression %s is enabled", self.compression)

    def is_alive(self, timeout=3, raise_error=False) -> bool:
        try:
            response = self.connect.get(self.tsdb_urls.version, timeout=timeout)
            if response.status_code != 200:
                raise ConnectionError("Bad status code")
            return True
        except Exception as error:
            if raise_error:
                raise TSDBNotAlive(str(error))
            return False

    @property
    def connect(self) -> Session:
        if not self._connect:
            self._connect = Session()
            if self.compression:
                self._connect.headers.update({'Content-Encoding': self.compression})
        return self._connect

    def sendall(self, *metrics) -> dict:
        logger.debug("Send metrics:\n %s", '\n'.join(str(m) for m in metrics))
        response = self.connect.post(
            self.tsdb_urls.put,
            data=gzip.compress(json.dumps(metrics).encode()) if self.compression else json.dumps(metrics),
            timeout=self.SEND_TIMEOUT)
        return response.json()



from opentsdb.protocols.http_connect import TSDBUrls


def test_from_host_port():
    url = TSDBUrls.from_host_and_port('127.0.0.1', 4242, version_endpoint="/api/version",
                                      put_endpoint="/api/put?details=true")
    assert url.put == 'http://127.0.0.1:4242/api/put?details=true'


def test_from_uri():
    url = TSDBUrls.from_uri('https://127.0.0.1:4242/opentsdb', version_endpoint="/api/version",
                            put_endpoint="/api/put?details=true")
    assert url.version == 'https://127.0.0.1:4242/opentsdb/api/version'





class TSDBClientException(Exception):
    pass


class TSDBNotAlive(TSDBClientException):
    pass


class TagsError(TSDBClientException):
    pass


class ValidationError(TSDBClientException):
    pass


class UnknownTSDBConnectProtocol(TSDBClientException):
    def __init__(self, protocol):
        self.protocol = protocol

    def __str__(self):
        return "Unknown TSDB connection protocol: %s" % self.protocol



from opentsdb.protocols.http_connect import HttpTSDBConnect
from opentsdb.protocols.telnet_connect import TelnetTSDBConnect
from opentsdb.push_thread import HTTPPushThread, TelnetPushThread
from opentsdb.exceptions import UnknownTSDBConnectProtocol

__all__ = ['HttpTSDBConnect', 'TelnetTSDBConnect', 'TSDBConnectProtocols']


class TSDBConnectProtocols:
    HTTP = 'HTTP'
    TELNET = 'TELNET'

    @classmethod
    def get_connect(cls, protocol: str, *args, **kwargs):
        if protocol == cls.HTTP:
            return HttpTSDBConnect(*args, **kwargs)
        elif protocol == cls.TELNET:
            return TelnetTSDBConnect(*args, **kwargs)
        raise UnknownTSDBConnectProtocol(protocol)

    @classmethod
    def get_push_thread(cls, protocol, *args, **kwargs):
        if protocol == cls.HTTP:
            return HTTPPushThread(*args, **kwargs)
        elif protocol == cls.TELNET:
            return TelnetPushThread(*args, **kwargs)
        raise UnknownTSDBConnectProtocol(protocol)
